
#include "Class.h"
using namespace std;

Class::Class(string Name, Class* BaseClass)
    : Type(Name)
{
    this->BaseClass = BaseClass;
    if(BaseClass != NULL)
        BaseClass->SubClasses.push_back(this);
}

Class::~Class()
{
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        delete *i;
    Members.clear();
}

void Class::Read(istream& in)
{
    string token;
    
    for(in.peek(); !in.eof(); in.peek())
    {
        in >> token;
        if(token == "{")
            break; // continue with reading the members
    
        if(token == ":")
        {
            in >> token;
            if(Type::AllTypes.find(token) != AllTypes.end())
            {
                Class* asClass = dynamic_cast<Class*>(AllTypes[token]);
                if(asClass != NULL)
                {
                    this->BaseClass = asClass;
                    BaseClass->SubClasses.push_back(this);
                }
                else
                    cerr << "Cannot derive from Primitive Type \"" << token << "\"." << endl;
            }
            else
                cerr << "Base-Class \"" << token << "\" not found." << endl;
        }
    }
    
    for(in.peek(); !in.eof(); in.peek())
    {
        in >> token;
        if(token == "};")
            return;
        
        if(Type::AllTypes.find(token) != AllTypes.end())
        {
            Property* newProperty = new Property(AllTypes[token], "");
            newProperty->Read(in);
            this->Members.push_back(newProperty);
        }
        else
            cerr << "Type-Name \"" << token << "\" not found." << endl;
    }
}





void Class::WriteTypeDefinition(ostream& os)
{
    os << "class " << TypeName << " extends " << (BaseClass != NULL ? BaseClass->TypeName : "Object") << endl;
    os << "{" << endl;

    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteDeclaration(os);

    os << endl << "    public function __construct()" << endl;
    os << "    {" << endl;
    os << "        parent::__construct();" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteConstructorDefinition(os);
    os << "    }" << endl;

    // RPC-Stub
    os << endl << "    public function RemoteProcedureCall($prefix, $str)" << endl;
    os << "    {" << endl;
    os << "        if(parent::RemoteProcedureCall($prefix, $str))" << endl;
    os << "            return true;" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
    {
        if((*i)->Array)
        {
            os << "        if($str == $prefix . '[" << (*i)->Name << "].Add()')" << endl;
            os << "            {array_push($this->" << (*i)->Name << ", new " << (*i)->type->TypeName << "()); return true;}" << endl;
        }
        if(dynamic_cast<PrimitiveType*>((*i)->type) == NULL)
        {
            if((*i)->Array)
            {
                os << "        $" << (*i)->Name << "Counter = 0;" << endl;
                os << "        foreach($this->" << (*i)->Name << " as &$" << (*i)->Name << "Iterator)" << endl;
                os << "        {" << endl;
                os << "            $" << (*i)->Name << "Selector = $prefix . '[" << (*i)->Name << "][' . $" << (*i)->Name << "Counter . ']';" << endl;
                os << "            if(substr($str, strlen($" << (*i)->Name << "Selector)) == $" << (*i)->Name << "Selector)" << endl;
                os << "                return $" << (*i)->Name << "Iterator->RemoteProcedureCall($" << (*i)->Name << "Selector, $str);" << endl;
                os << "            $" << (*i)->Name << "Counter++;" << endl;
                os << "        }" << endl;
            }
            else
            {
                os << "        if(substr($str, strlen($prefix)+" << (*i)->Name.length()+2 << ") == $prefix . '[" << (*i)->Name << "]')" << endl;
                os << "            return $this->" << (*i)->Name << "->RemoteProcedureCall($prefix . '[" << (*i)->Name << "]', $str);" << endl;
            }
        }
    }
    os << "        return false;" << endl;
    os << "    }" << endl;
    
    
    WriteXmlLoader(os);
    WriteXmlSaver(os);
    WriteHtmlShow(os);
    WriteHtmlForm(os);
    WriteHttpPostLoader(os);

    os << "}" << endl;
}
void Class::WriteConstructorDefinition(Property* property, ostream& os)
{
    if(property->Array)
        os << "        $this->" << property->Name << " = array();" << endl;
    else
        os << "        $this->" << property->Name << " = new " << TypeName << "();" << endl;
}




void Class::WriteXmlLoader(ostream& os)
{
    os << endl << "    public function fromXml(&$xml)" << endl;
    os << "    {" << endl;
    os << "        parent::fromXml($xml);" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteXmlLoadCall("xml", os);
    os << "    }" << endl;
}
void Class::WriteXmlLoadCall(string xmlName, Property* property, ostream& os)
{
    if(property->Array)
    {
        os << "        foreach($" << xmlName << "->" << property->Name << " as $" << property->Name << "XmlNode)" << endl;
        os << "        {" << endl;
        os << "            $" << property->Name << "TypeName = (string)$" << property->Name << "XmlNode['type'];" << endl;
        os << "            $" << property->Name << "Temp = new $" << property->Name << "TypeName;" << endl;
        os << "            $" << property->Name << "Temp->fromXml($" << property->Name << "XmlNode);" << endl;
        os << "            array_push($this->" << property->Name << ", $" << property->Name << "Temp);" << endl;
        os << "        }" << endl;
    }
    else
    {
        os << "        $" << property->Name << "XmlNode = $" << xmlName << "->" << property->Name << ";" << endl;
        os << "        $" << property->Name << "TypeName = (string)$" << property->Name << "XmlNode['type'];" << endl;
        os << "        $this->" << property->Name << " = new $" << property->Name << "TypeName;" << endl;
        os << "        $this->" << property->Name << "->fromXml($" << property->Name << "XmlNode);" << endl;
    }
}




void Class::WriteXmlSaver(ostream& os)
{
    os << endl << "    public function toXml(&$xml)" << endl;
    os << "    {" << endl;
    os << "        parent::toXml($xml);" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteXmlSaveCall("xml", os);
    os << "    }" << endl;
}
void Class::WriteXmlSaveCall(string xmlName, Property* property, ostream& os)
{
    if(property->Array)
    {
        os << "        foreach($this->" << property->Name << " as $" << property->Name << "Iterator)" << endl;
        os << "        {" << endl;
        os << "            $" << property->Name << "XmlNode = $" << xmlName << "->addChild('" << property->Name << "');" << endl;
        os << "            $" << property->Name << "Iterator->toXml($" << property->Name << "XmlNode);" << endl;
        os << "        }" << endl;
    }
    else
    {
        os << "        $" << property->Name << "XmlNode = $" << xmlName << "->addChild('" << property->Name << "');" << endl;
        os << "        $this->" << property->Name << "->toXml($" << property->Name << "XmlNode);" << endl;
    }
}




void Class::WriteHtmlShow(ostream& os)
{
    os << endl << "    public function htmlView($indent, $divid)" << endl;
    os << "    {" << endl;
    os << "        if($divid != null)" << endl;
    os << "        {" << endl;
    os << "            echo str_repeat('  ', $indent) . \"<div class='composed'><div class='" << TypeName << "' id='$divid'>\";" << endl;
    os << "            echo \"<div class='caption'>$divid</div>\\n\";" << endl;
    os << "        }" << endl;
    os << "        parent::htmlView($indent, null);" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteHtmlShowCall(os);
    os << "        if($divid != null)" << endl;
    os << "            echo str_repeat('  ', $indent) . \"</div></div>\\n\";" << endl;
    os << "    }" << endl;
}
void Class::WriteHtmlShowCall(Property* property, ostream& os)
{
    if(property->Array)
    {
        os << "        echo str_repeat('  ', $indent+1) . \"<div class='array' id='" << property->Name << "'><div class='caption'>" << property->Caption << "</div>\\n\";" << endl;
        os << "        $" << property->Name << "Counter = 0;" << endl;
        os << "        foreach($this->" << property->Name << " as $" << property->Name << "Iterator)" << endl;
        os << "            $" << property->Name << "Iterator->htmlView($indent+2, '" << property->Name << "');" << endl;
        os << "        echo str_repeat('  ', $indent+1) . \"</div>\\n\";" << endl;
    }
    else
    {
        os << "        $this->" << property->Name << "->htmlView($indent+1, '" << property->Name << "');" << endl;
    }
}




void Class::WriteHtmlForm(ostream& os)
{
    os << endl << "    public function htmlForm($prefix, $indent, $divid)" << endl;
    os << "    {" << endl;
    os << "        if($divid != null)" << endl;
    os << "        {" << endl;
    os << "            echo str_repeat('  ', $indent) . \"<div class='composed'><div class='" << TypeName << "' id='$divid'>\\n\";" << endl;
    os << "            echo str_repeat('  ', $indent+1) . \"<div class='caption'>$divid</div>\\n\";" << endl;
    os << "        }" << endl;
    os << "        parent::htmlForm($prefix, $indent, null);" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteHtmlFormCall(os);
    os << "        if($divid != null)" << endl;
    os << "            echo str_repeat('  ', $indent) . \"</div></div>\\n\";" << endl;
    os << "    }" << endl;
}
void Class::WriteHtmlFormCall(Property* property, ostream& os)
{
    if(property->Array)
    {
        os << "        echo str_repeat('  ', $indent+1) . \"<div class='array' id='" << property->Name << "'><div class='caption'>" << property->Caption << "</div>\\n\";" << endl;
        os << "        $" << property->Name << "Counter = 0;" << endl;
        os << "        foreach($this->" << property->Name << " as $" << property->Name << "Iterator)" << endl;
        os << "            $" << property->Name << "Iterator->htmlForm($prefix . '[" << property->Name << "][' . $" << property->Name << "Counter++ . ']' , $indent+2, '" << property->Name << "');" << endl;

        os << "        echo str_repeat('  ', $indent+2) . \"<button type='submit' name='rpc' value='\" . $prefix . \"[" << property->Name << "].Add()'>Add</button>\\n\";" << endl;
        os << "        echo str_repeat('  ', $indent+1) . \"</div>\\n\";" << endl;
    }
    else
    {
        os << "        $this->" << property->Name << "->htmlForm($prefix . '[" << property->Name << "]' , $indent+1, '" << property->Name << "');" << endl;
    }
}




void Class::WriteHttpPostLoader(ostream& os)
{
    os << endl << "    public function fromHttpPost(&$post)" << endl;
    os << "    {" << endl;
    os << "        parent::fromHttpPost($post);" << endl;
    for(vector<Property*>::iterator i = Members.begin(); i != Members.end(); i++)
        (*i)->WriteHttpPostLoadCall(os);
    os << "    }" << endl;
}
void Class::WriteHttpPostLoadCall(Property* property, ostream& os)
{
    if(property->Array)
    {
        os << "        $" << property->Name << "Counter = 0;" << endl;
        os << "        foreach($this->" << property->Name << " as $" << property->Name << "Iterator)" << endl;
        os << "            $" << property->Name << "Iterator->fromHttpPost($post['" << property->Name << "'][$" << property->Name << "Counter++]);" << endl;
        
        /*
        os << "        if($post['" << property->Name << "']['NewElementAdded'])" << endl;
        os << "        $" << property->Name << "NewItem = new " << property->type->TypeName << "();\n";
        os << "        $" << property->Name << "NewItem->fromHttp($post[" << property->Name << "][' . $" << property->Name << "Counter++ . ']);" << endl;
        os << "        array_push($this->" << property->Name << ", $" << property->Name << "NewItem);" << endl;
        */
    }
    else
    {
        os << "        $this->" << property->Name << "->fromHttpPost($post['" << property->Name << "']);" << endl;
    }
}